import binascii
import psycopg2
from psycopg2.extensions import new_type, register_type

try:
    import cPickles as pickle
except ImportError:
    import pickle


def adapt_bytea(obj):
    '''Adapt an object to a bytea by pickling.'''
    p = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)
    return psycopg2.Binary(p)


def cast_bytea(value, cur):
    '''Convert a bytea to a python value by unpickling.'''

    # Decode the bytea using the original typecast object.
    value = psycopg2.BINARY(value, cur)
    try:
        return pickle.loads(value)
    except pickle.UnpicklingError:
        mesg = 'unable to unpickle buffer: {!r}'.format(value)
        raise psycopg2.InterfaceError(mesg)


# Register cast_bytea with psycopg2.
PICKLE = new_type(psycopg2.BINARY.values, 'PICKLE', cast_bytea)
register_type(PICKLE)


def dummy_pool(conn):
    class DummyPool(object):
        def getconn(self):
            return conn
        def putconn(self, conn):
            pass
    return DummyPool()


class Garden(object):
    '''A place to store your cucumbers.'''

    def_fmt = '''
    CREATE TABLE {name}
        ( key   bytea     NOT NULL UNIQUE
        , value bytea     NOT NULL
        , mtime timestamp NOT NULL DEFAULT localtimestamp
        )
    '''

    replace_fmt = '''
    CREATE RULE "replace_{name}" AS
        ON INSERT TO "{name}"
        WHERE
            EXISTS(SELECT 1 FROM {name} WHERE key=NEW.key)
        DO INSTEAD
            UPDATE {name}
            SET value = NEW.value, mtime = localtimestamp
            WHERE key = NEW.key
    '''

    select_fmt = 'SELECT value FROM {name} WHERE key = %s'
    insert_fmt = 'INSERT INTO {name} (key, value) VALUES (%s, %s)'

    def __init__(self, name, pool):
        self.name = name
        self.pool = pool

        # Format the class name into the table and rule definitions.
        self.table_def = self.def_fmt.format(name=self.name)
        self.replace_cmd = self.replace_fmt.format(name=self.name)
        self.select_cmd = self.select_fmt.format(name=self.name)
        self.insert_cmd = self.insert_fmt.format(name=self.name)

        conn = self.pool.getconn()

        # Create the table and replacement rule if not already defined.
        with conn.cursor() as cur:

            cur.execute('''
                SELECT 1 FROM pg_tables WHERE tablename = '{name}'
                '''.format(name=self.name))
            if not cur.fetchone():
                cur.execute(self.table_def)

            cur.execute('''
                SELECT 1 FROM pg_rules WHERE rulename = 'replace_{name}'
                '''.format(name=self.name))
            if not cur.fetchone():
                cur.execute(self.replace_cmd)

        conn.commit()
        self.pool.putconn(conn)

    def __getitem__(self, key):
        '''Retrieve a cucumber from the Garden.'''

        key = adapt_bytea(key)

        conn = self.pool.getconn()
        with conn.cursor() as cur:
            cur.execute(self.select_cmd, (key,))
            value = cur.fetchone()
        self.pool.putconn(conn)

        if value is not None:
            value = value[0]
        return value

    def __setitem__(self, key, value):
        '''Place/replace a cucumber into the Garden.'''

        key = adapt_bytea(key)
        value = adapt_bytea(value)

        conn = self.pool.getconn()
        with conn.cursor() as cur:
            cur.execute(self.insert_cmd, (key, value))
        conn.commit()
        self.pool.putconn(conn)



import collections
import inspect


def cucumber(typename, field_names, verbose=False, rename=False,
             version=None, migrations=None):
    '''Like namedtuple, but better for making pickles.'''

    if migrations is None:
        migrations = {}

    # Convert field_names early to detect length.
    if isinstance(field_names, basestring):
        field_names = field_names.replace(',', '').split()

    # Require at least two fields to avoid confusion during unpickling.
    if len(field_names) == 1:
        raise ValueError, 'cucumbers must have at least 2 fields'

    # Start out with a namedtuple.
    cls = collections.namedtuple(typename, field_names, verbose, rename)

    def __getnewargs__(self):
        '''A singleton tuple containing the version and arguments.
        
           This overrides namedtuple's __getnewargs__().
        '''
        return (tuple([self._version] + list(self)),)

    def __new__(cls, *args, **kwargs):
        '''{cls.__new__.__doc__}

           When only one argument is provided, use that argument as
           the state for restoring from a pickled representation.
        '''.format(cls=cls)

        # Cucumbers with only one field are forbidden to avoid confusion here.
        if len(args) == 1 and not kwargs:
            version, args = args[0][0], args[0][1:]
            # Perform a migration if needed.
            if version != cls._version:
                args = cls._migrations[version, cls._version](*args)

        # Figure out args and kwargs.
        args = list(args)
        for field in cls._fields[len(args):]:
            if field not in kwargs:
                raise TypeError, 'expecting argument {!r}'.format(field)
            args.append(kwargs.pop(field))

        if kwargs:
            args = ', '.join(repr(name) for name in kwargs)
            raise TypeError, 'unexpected arguments: {!s}'.format(args)

        return tuple.__new__(cls, args)

    def __getstate__(self):
        '''Minimize pickle size by excluding state.'''

    def __setstate__(self):
        '''Minimize pickle size by excluding state.'''

    def migrate_from(cls, old_version, new_version):
        '''Decorator for defining migrations between versions.'''
        def migration(func):
            cls._add_migration(old_version, new_version, func)
            return func
        return migration

    def _add_migration(cls, old_version, new_version, func):
        '''Add a migration.'''
        for (old, new), other_func in cls._migrations.items():
            if new == old_version:
                composed = lambda *args: func(*other_func(*args))
                cls._add_migration(old, new_version, composed)
        cls._migrations[old_version, new_version] = func

    # Construct some source code for verbose outcode.
    if verbose:

        # Try to make verbose source code as correct as possible by
        # by removing the string formatting I used above.
        repr_new = inspect.getsource(__new__) \
                    .replace('.format(cls=cls)', '') \
                    .replace('{cls.__new__.__doc__}', cls.__new__.__doc__)

        print '    # The following have been added by cucumber:'
        print
        print '    _version =', repr(version)
        print '    _migrations = {}'
        print
        print inspect.getsource(__getnewargs__)
        print repr_new
        print '    @classmethod'
        print inspect.getsource(migrate_from)
        print '    @classmethod'
        print inspect.getsource(_add_migration)
        print inspect.getsource(__getstate__)
        print inspect.getsource(__setstate__)
        for k, v in migrations.items():
            fmt = '{!s}._add_migration({!r}, {!r}, {.__name__!s})'
            print fmt.format(typename, k[0], k[1], v)

    # Add new fields to the class.
    cls._version       = version
    cls._migrations    = {}
    cls.__getnewargs__ = __getnewargs__
    cls.__new__        = staticmethod(__new__)
    cls.__getstate__   = __getstate__
    cls.__setstate__   = __setstate__
    cls.migrate_from   = classmethod(migrate_from)
    cls._add_migration = classmethod(_add_migration)
    for k, v in migrations.items():
        cls._add_migration(k[0], k[1], v)

    # Correct __module__ for pickling to work correctly. 
    try:
        frame = inspect.currentframe().f_back
        cls.__module__ = frame.f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return cls



from distutils.core import setup

with open('version.txt') as f:
    version = f.read().strip()

url = 'https://github.com/nejstastnejsistene/gardendb'

classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Database',
    ]

setup(
    name = 'gardendb',
    packages = ['gardendb'],
    version = version,
    description = 'Simple python flat-file databases with Cucumbers: '
                  'simple, version-controlled records with streamlined '
                  'pickle representations.',
    author = 'Peter Johnson',
    author_email = 'pajohnson@email.wm.edu',
    url = url,
    download_url = url + '/releases',
    keywords = [],
    classifiers = classifiers,
    long_description = '''\
This is a WIP, see the readme for more info.
'''
    )
